"""

BrahmA Evolution Module
=======================================
written by Rahul Tandon c/o Vixen Intelligence

"""

import os, sys, time, random, copy


#--config core import--
# sys.path.append('../')
# from configCore import *

#import root bot
from marlin_brahma.bots.bot_root import *

# import bots
from custom_bots import *


#-----------------------------MATE-----------------------------------------------


class Mate(object):
  '''
  Class to mate individuals. Static and class routines provided to mate bots. This is 
  the default version.

  Procedure:
  1. Create a child bot
  2. Combine parent DNA
  3. Add randomly chosen tradnscription DNA from parents.
  3. Return child

  '''

  def __init__(self):
    pass

  @staticmethod
  def Mate(maleBot, femaleBot):

    
    active_species = maleBot.species
    active_env = femaleBot.env
    direction = femaleBot.direction
    child_args = {'direction' : direction, 'parent' : femaleBot.name}
    botStr = eval(active_species)
    childBot = botStr(active_env, myspecies = active_species, myargs = child_args)
    
    mom_tmp = copy.deepcopy(femaleBot)
    dad_tmp = copy.deepcopy(maleBot)

    #determine number of dna strands
    import math
    numberMaleDNA = math.floor(maleBot.numberDNA)
    numberMaleDNA = max(1, numberMaleDNA)
    numberFemaleDNA = math.floor(femaleBot.numberDNA)
    numberFemaleDNA = max(1, numberFemaleDNA)

    #determine which DNA strands

    #forward or reverse

    #addDNAStrand(self, dna):
    for i in range(numberMaleDNA):
        # print (f'number genes: {len(dad_tmp.dNA)}')
        if len(dad_tmp.dNA) < 1:
          continue
        dnaTag = random.choice(list(dad_tmp.dNA.keys()))
        Mate.addDNAStrand(dad_tmp.dNA[dnaTag], childBot)
        del dad_tmp.dNA[dnaTag]

    #forward or reverse
    
    # for i in range(numberFemaleDNA):
    #     print (f'number genes: {len(mom_tmp.dNA)}')
    #     if len(mom_tmp.dNA) < 1:
    #       continue
    #     dnaTag = random.choice(list(mom_tmp.dNA.keys()))
    #     Mate.addDNAStrand(mom_tmp.dNA[dnaTag], childBot)
    #     del mom_tmp.dNA[dnaTag]

    #transcription DNA
    Mate.addTranscription(random.choice([femaleBot.transcriptionDNA,maleBot.transcriptionDNA]),  childBot)

    
   
    
    del mom_tmp
    del dad_tmp

    #return child
    return childBot

  """
  Evolutionary procedures. These can be overritten, but these are provided 
  by BrahmA in the root bot class.
  """

  @staticmethod
  def addTranscription(transcription, patient):
    '''
    Add a transcription to a child bot  (or possibkly adult)

    Arguments:
        transcription {Transcription} -- Protein transcription algorithm.

    '''
    newTanscription = copy.deepcopy(transcription)
    patient.transcriptionDNA = newTanscription
    
    return patient

  @staticmethod
  def addDNAStrand(dna, patient):
    '''Copy DNAStrand to bot

    Arguments:
        dna {[VixenDNA]} -- DNAStrand
    '''

    patient.numberDNA+=1
    newDNA = copy.deepcopy(dna)
    newDNATag = random.randint(100,100000)
    newDNA.Name = newDNATag

    patient.dNA[newDNATag] = newDNA
    patient.dNAExpression[newDNATag] = 0.0
    

    return patient

  @staticmethod
  def removeDNAStrand(patient):

    if len(patient.dNA > 1):
          #get key from dnaStrands and del from bot
        randomKey = random.choice(list(patient.dNA.keys()))
        del patient.dNA[randomKey]
        del patient.dNAExpression[randomKey]
    
    return patient
    


#--------------------------TOURNAMENT--------------------------------------------


class RootTournament(object):
  def __init__(self, generationEval = None, population = None, dataManager = None):
    self.population = population
    #ranking by name
    self.rankings = []
    #evaluations
    self.evaluations = generationEval
    #optimisation data manager

    if (len(self.population.bots) < 4):
      print ("Critical Error! Need more bots. Exiting.")
      exit()

  def Rank(self):
    pass

  def Regenerate(self):
    pass

class SimpleTournamentRegenerate(RootTournament):
  """
  A simple tournament (ranking algorithm) shipped with BrahmA for rapid develpment.s
  This class 

  :param RootTournament: [description]
  :type RootTournament: [type]
  """

  def __init__(self, generationEval = None, population = None, dataManager = None):
    super().__init__(generationEval = generationEval, population = population, dataManager = dataManager)
    if generationEval != None:
        #evaluations (EvaluateDecisions) by bot Tag
        #self.evaluations = generationEval
        #need the population data as this class will kill and create new population members
        #edited Nov 2020. Taken care in root
        #self.population = population
        #ranking by name
        #self.rankings = []
        pass

    else:
        print ("critical error initialising tournament.")
        exit()


  def Zeros(self):
    zeros = []

    for k, v in self.evaluations.items():
      if v.fitnessValue == 0.0 or v.fitnessValue == -1000:
        #we have no decisions, return
        zeros.append(k)

    number_zeros = len(zeros)
    # print (f'Number of dead wood: {number_zeros}')
    return zeros


  def RankPopulation(self, output = 0 ):

    import operator
    # print ("evaluations")
    # print (self.evaluations)
   
    rankedEvals = sorted(self.evaluations.items(), key=lambda x: x[1].fitnessValue, reverse=True)
    # print ("Ranked")
    #print (rankedEvals)
    #reset rankings
    
    self.rankings = []
    for key, value in rankedEvals:
        
        # print (f'val: {value.fitnessValue}')
        if value.fitnessValue != 0.0 or value.fitnessValue != 0:
          self.rankings.append(key)
          # print (f'{key} : {value.fitnessValue}')
          # print (f'added: {value.fitnessValue} ')
        #print (botID)
    # print (self.rankings)
    if len(self.rankings) < 5:
      self.rankings = []
      #n = len(self.rankings)
      for key, value in rankedEvals:
        #print (f'val: {value.fitnessValue}')
        self.rankings.append(key)
        #print (f'added: {value.fitnessValue} ')

    # print (self.rankings)
 

    if output == -1:
      winners = []
      for botID, value in self.evaluations.items():
        winners.append(botID)
      
      return winners


    if output == 1:
        import random

        winners = []
        unique_names = []
        for botID, value in self.evaluations.items():
            if value.fitnessValue > 0:

                uniqueName = botID + "_" + str(random.randint(10,1000000))
                unique_names.append(uniqueName)
                winners.append(botID)

            #logger.info('Optimisation: %d', generation)

        return winners
        
  def RegeneratePopulation(self):

    #select 2 parents...

    #best bot structure 1
    maleParent = self.population.species[self.rankings[0]]
    # print (f'parent : {maleParent.name}')
    #best bot structure 2
    femaleParent = self.population.species[self.rankings[1]]
    # print (f'parent : {femaleParent.name}')


    
    
    #mom_tmp = copy.deepcopy(femaleParent)
    #dad_tmp = copy.deepcopy(maleParent)
    childOne = Mate.Mate(maleParent, femaleParent)

    #best bot structure 3
    maleParent = self.population.species[self.rankings[2]]
    # print (f'parent : {maleParent.name}')
    
    #best bot structure 4
    femaleParent = self.population.species[self.rankings[3]]
    # print (f'parent : {femaleParent.name}')

    childTwo = Mate.Mate(maleParent, femaleParent)

    #kill bottom 2
    populationSize = len(self.population.species)
    #edit rankings as ranking vevtor now not as long as population  as zero decision makers removed
    deathTagOne = self.rankings[len(self.rankings)-1]
    deathTagTwo = self.rankings[len(self.rankings)-2]

    del (self.population.species[deathTagOne])
    
    self.population.bots.remove(deathTagOne)
    del (self.population.species[deathTagTwo])
    
    self.population.bots.remove(deathTagTwo)

    #add children
    self.population.species[childOne.name] = childOne
    self.population.bots.append(childOne.name)
    self.population.species[childTwo.name] = childTwo
    self.population.bots.append(childTwo.name)


    mom_tmp = None
    dad_tmp = None

#-------------------------- GENETIC SHUFFLE -------------------------------

class RootMutate(object):
    

  """
  Root Mutate Class
  =============================
  Individual (gene) mutation is taken care of the root bot class. The mutation routine
  can of course be overridden. 
  """



  def __init__(self,  population = None, config = None):
    self.population = population
    #edit nov 2020 - read directly from optimisation parameters
    self.config = config


  def mutate(self, args = {}):
    '''
    Mutate according to mutation rate.
    '''

    for indTag, ind in self.population.species.items():
        chance = random.random()
        if chance < self.config['mutation_rate']:
            #print (f'Mutating...[{indTag}]')
            ind.Mutate(args)



#========================================================================
#                 Evolve                                                =
#========================================================================
class Evolve(object):
  """
  Main callable evolve class for BrahmA
  """
  def __init__(self):
    pass

#=======================================================================


#========================================================================
#                 Evolve Data                                               =
#========================================================================
class EvolveData(object):
  """
  Main callable evolve class for BrahmA
  """
  def __init__(self):
    pass

#=======================================================================


"""This is v1.0 of the trader template. Procedures for birth, death evolution must be contained here.
This is a template for a general intraday futures/fx trader.
"""

#v2.0 
import os, sys, time, copy, datetime, random

#config core
# sys.path.append('../')
# from configCore import *

# sys.path.append('../dna')
# current
import logging
# importing
from marlin_brahma.dna.vixen_dna import *
from marlin_brahma.transcription.transcription import * 


class BotRoot(object):
    
    """BotRoot is the root class of all indis/bots in the brahma world.
    Genetic material is deterined by lcl folder for material and derived bots  can add
    functionality. e.g.  trader  - direction of bot.

    todo: visualise bot genetic strucure

    :param object: [description]
    :type object: [type]
    :param object: [description]
    :type object: [type]
    :param object: [description]
    :type object: [type]

    """
    def __init__(self, myenv="", myspecies = "BotRoot", myargs=None, version="1_0_0"):
        self.direction = 0
        self.env = myenv
        self.dob = datetime.datetime.now()
        self.species = myspecies
        self.custom = False #introduced for custom traders built elsewhere
        self.name = "vixen_bot" + str(random.randint(0,1000000))  + str(random.randint(0,1000000)) 
        self.version = version
        self.regulatory_network = None
        #genetic material
        if myargs == None:
            
            self.parent = "Eve"
        else:
            if 'parent' in myargs:
                self.parent = myargs['parent']
            if 'custom' in myargs:
                self.custom = myargs['custom']
            if 'name' in myargs:
                self.name = myargs['name']
        
        self.dNA = {}           #contains strands of DNA
        #self.riskDNA = {}
        self.numberDNA = 0
        self.dNAExpression = {}         #expression value of each DNA strand 
                                        #this is required for transcription

        #self.dNARiskExpression = {}     #expression value of each risk DNA strand 
                                        #this is required for transcription

        #protein transcription
        self.transcriptionDNA = None    #this is what is used for transcription
                                        #DNA expression levels is required

        #initialised
        self.initialised = False

        #logger
        #self.logger = GetBrahmaLogger("Bot Creation")
        

    def __str__(self):
        """
        Overide of print method.Export high level description of bot. 

        :return: High  level description of bot.
        :rtype: String
        """
        risk_tag = "No Risk DNA"
        num_risk = 0
        for tag, dna in self.dNA.items(): 
            if dna.dnaType == "risk":
                risk_tag = dna.name
                num_risk += 1
                
        
        if not hasattr(self, 'version'):
            self.version = "1_0_0"
        
        return ("Name: [{0}] Version [{7}] Custom: [{6}] Number DNA: [{1}] Env: [{2}] Species: [{3}] Risk Tag: [{4}] Risk DNA[{5}]".format(self.name, self.numberDNA, self.env, self.species, risk_tag, num_risk, self.custom, self.version))

    def BuildBot(self, parms = None, gene_args = None):
        """
        Build the bot and all its genetic material. This is the root class 
        so all root genetic material is built. 

        :param parms: [description], defaults to None
        :type parms: [type], optional
        """
        #print ("here")
        #self.logger.info("***Building Bot : {0}".format(self.name))
        # print (f'building version  : {self.version}')
        #determine numer of dna strands
        if parms != None:
            self.numberDNA = random.randint(parms['min_number_dna'], parms['max_number_dna'])
            logging.debug(f'built dna number')
        else:
            #Build generic bot.
            exit(0)



        #build risk DNA - alpha - one for now.  keep simple.
        #riskDNA =  VixenDNA(deltaD = 1, dataLength=1, envName=self.env, dataDelay=1)
        #riskDNA_tag = riskDNA.name
        #riskDNA.AddRiskGenome()
        #self.riskDNA[riskDNA_tag] = riskDNA
        #self.dNARiskExpression[riskDNA_tag] = 0.0
        #del(riskDNA)
        
        
        #--------------------------- Initial MyTradeBot
        
        #number of marketDNA
        for i in range(0,self.numberDNA):
           #print (self.numberDNA)
            
            #build market dna

            #get deltaT
            #dna_deltaT = random.randint(1, parms['maxDeltaT'])
            #dna_deltaT = random.choice(EpochBuckets)
            
            #get length
            dna_length = random.randint(parms['min_number_genes'], parms['max_number_genes'])
            #get delay
            dna_delay = random.randint(0, parms['max_dna_data_delay'])
            
            # brahma note: group data on fly
            logging.debug(f'dna adding dna structure')
            botDNA = VixenDNA(dataLength=dna_length, envName=self.env, dataDelay=dna_delay, gene_args = gene_args)
            botDNA_tag = botDNA.name
           
            #add genome... only one per DNA strand - yes for now
            logging.debug(f'dna adding genome')
            botDNA.AddGenome(parms=parms)
            
            # print ("dna: {}".format(i))
            # ()exit
            
            self.dNA[botDNA_tag] = botDNA
            
            # del(marketDNA)
            botDNA = None
            self.dNAExpression[botDNA_tag] = 0.0
            #print (self.MarketDNA[marketDNA_tag].ExpressionTable)
            # exit()

            logging.debug(f'built genome')
        
        

        #build transcription DNA
        self.transcriptionDNA = Transcription()
         
    def BuildChild(self, parent = None):
        species = parent.species

    def Reset(self):
        for dnaTag, dna in self.dNA.items():
            dna.Reset()
            
    def GetMemory(self):
        max_memory = -1
        for dnaTag, dna in self.dNA.items():
            max_memory = max(max_memory, dna.GetMemory())
            # print (f'b {max_memory}')
            
        # print (f'bot {max_memory} ')    
        return max_memory

    def StartUp(self):
        """
        All expression levels must be set to zero before starting a new run.
        """

        for dnaTag, e in self.dNAExpression.items():
            self.dNAExpression[dnaTag] = 0.0

        #for dnaTag, e in self.dNARiskExpression.items():
        #    self.dNARiskExpression[dnaTag] = 0.0
    
    def ForceTranscribe(self):
        for dnaTag, expression in self.dNAExpression.items():
            self.dNAExpression[dnaTag] = 1
    
    def ExpressDNA(self, data={}):
        """
        ---Populate dNAExpression---
        Express each strand of DNA in the bot and build the DNA epression table for
        the bot. Return values are 1 for bot is now initialised and 0 for not initialised.

        :param data: Pressure data, defaults to {} for root as no data required 
                    for test data. User defined for derived bot data requirements
        :type data: dict, optional
        :return: Initialised state of bot. True: is initialised False: not initialised 
        :rtype: Bool
        """

        #run risk first - obsolete
        '''
        for dnaTag, dna in self.riskDNA.items():
            expression_value = dna.ExpressDNA(data=data)
            #check to see if all the genetic material has been initialised
            #a value of False means that one of the genomes has not been initialised
            #so ExpressDNA also returns false after settng the initialised value to  False
            if expression_value == False:
                self.initialised = False
                return  False
            self.dNARiskExpression[dnaTag] = expression_value
        '''
        
        #added feb 21. Reset expression between iterations
        #self.dNAExpression  = {}
        #--use StartUp instead

        #print (self.initialised)
        #print (self.dNA)
        #print (self.dNAExpression)
        
        
        for dnaTag, dna in self.dNA.items():
            # print (f'dna: {dnaTag}')
           
            
            pressure_data = data
            
            expression_value = dna.ExpressDNA(data=pressure_data)
            
            if dna.dnaType == "risk":
                if expression_value == 1:
                    #transcribe protein
                    self.ForceTranscribe()
                    return True
            
            
            #print  (f'******* {expression_value}')

            #debug ---> gene flip analysis
            #print  (expression_value)
            #check to see if all the genetic material has been initialised
            #a value of False means that one of the genomes has not been initialised
            #so ExpressDNA also returns false after settng the initialised value to  False


            #update nov 2020 :  returning boolean and floating point
            if expression_value == -10:
                self.initialised = False
                if dnaTag in self.dNAExpression:
                    self.dNAExpression[dnaTag] = 0.0
                else:
                    self.dNAExpression[dnaTag] = 0.0
                    print (self.dNAExpression)
                    print (dnaTag)
                    print ("Key error location 2. bot_root")
                    
                #feb 2021 
                return  False

            #each DNA strand in bot now has valie
            self.dNAExpression[dnaTag] = expression_value


        #we have expressed all DNA strands which means all genomes inside have been run
        #which means that the individual is now intialised.
        self.initialised = True

        #---debug output
        #tickExpress = []
        #for key, val in self.dNAExpression.items():
        #    #print (val)
        #    fp = open ("express.txt", "a")
        #    fp.write("{0}\n".format(val))
        #    fp.close()
        #    tickExpress.append(val)
        #print (self.initialised)
        #print (*tickExpress)
        
        #print ("***")
        #---

        return True

    def GetExpressionData(self, type=""):
        return self.dNAExpression
       
    def GetAvgExpressionValue(self):
        total_expression = 0.0
        num_dna = 0
        avg_expression = 0.0
        for k,v in self.dNAExpression.items():
            total_expression = total_expression + v
            num_dna+=1
            
        if num_dna > 0:
            avg_expression = total_expression/num_dna
        return avg_expression
            
    
    #get expression data for each chromosome / DNA for transcription -> one value per (MARKET)DNA STRAND
    def get_chromosome_expression_structure(self):
        data = {}

        for tagID, dna in self.dNA.items():

            if EXPRESS_DEBUG:
                dna.print_expression_table()

            sum_express = 0
            num_genome = 0
            num_genome = len(dna.genome)
            #print ("num gen ", num_genome)
            #print (marketDNA.ExpressionTable)
            for genomeID, genome in dna.genome.items():


                if genomeID in dna.expressionTable:
                    sum_express = sum_express + dna.expressionTable[genomeID]
                    if GENOME_DEBUG == True:
                        print ("Genome Expression Vector")
                        print (dna.expressionTable[genomeID])

                else:
                    if TRANSCRIPTION_DEBUG:
                        print ("Can't find genome expression for transcription data! Expression Tabe for chromosome: \n")
                        print (dna.expressionTable)

            chromo_express = sum_express/num_genome
            data[tagID]=chromo_express

            if TRANSCRIPTION_DEBUG:
                print ("Num Genome in Chromosome: " + str(num_genome))
                print ("Sum Express for DNA: {0}".format(sum_express))
                print ("Actual Express for DNA: {0}".format( chromo_express))





        return data

    def printMood(self):
        pass

    def printStr(self):
        import json
        """Expression and decision making output. Not just expression levels. Trader level debug.
        Get a feel for the mood.
        """
        #ßprint ("root bot print STr")
        #build the bot structure:
        bot_structure = {}
        bot_structure['name'] = self.name
        bot_structure['chromosomeNumber'] = self.numberDNA
        #bot_structure['direction'] = self.direction
        bot_structure['env'] = self.env
        if hasattr(self, 'version'):
            bot_structure['version'] = self.version
        else:
            bot_structure['version'] = '1_0_0'
            
        chromosomes = {}
        for dnaTag, chromosome in self.dNA.items():
            chromosomes[dnaTag] = chromosome.GetStructureData()

       
        bot_structure['genetics'] = chromosomes

        

        #export
        '''
        file_str = open(BOT_DATA_FOLDER + self.name + ".json", "w")
        json.dump(bot_structure, file_str)
        file_str.close()

        if CHILD_DEBUG:
            print (bot_structure)

        if STRUCTURE_OUTPUT == True:
            
            for dnaTag, chromosome in self.dNA.items():
                print (chromosome)
                for genomeID, genome in chromosome.genome.items():
                    print (genome)

       


        if MOOD_OUTPUT == True:

            for dnaTag, chromosome in self.dNA.items():
                print (chromosome)
                """
                if hasattr(self, 'MarketDNAExpression'):
                    print ("Excitement: {0}".format(self.MarketDNAExpression[dnaTag]))
                else:
                    chromosome.print_expression_table()"""
        '''
        return json.dumps(bot_structure)

    def printAll(self):
        """
        Print all ind genetic data
        """
        for key, dna_strand in self.dNA.items():
            print (dna_strand.__str__())
            for key, gene in dna_strand.genome.items():
                print (gene)
            
    
    """
    Evolutionary procedures. These can be overritten, but these are provided 
    by BrahmA in the root bot class.
    """
    """
    @staticmethod
    def addTranscription(transcription, patient):
        '''
        Add a transcription to a child bot  (or possibkly adult)

        Arguments:
            transcription {Transcription} -- Protein transcription algorithm.

        '''
        newTanscription = copy.deepcopy(transcription)
        patient.transcriptionDNA = newTanscription
        
        return patient

    @staticmethod
    def addDNAStrand(dna, patient):
        '''Copy DNAStrand to bot

        Arguments:
            dna {[VixenDNA]} -- DNAStrand
        '''

        patient.NumberMarketDNA+=1
        newDNA = copy.deepcopy(dna)
        newDNATag = random.randint(100,100000)
        newDNA.Name = newDNATag

        patient.marketDNA[newDNATag] = newDNA
        if DEBUG:
            print (patient.marketDNA[newDNATag])


        return patient

    @staticmethod
    def removeDNAStrand(patient):

        if len(patient.marketDNA > 1):
             #get key from dnaStrands and del from bot
            randomKey = random.choice(list(patient.marketDNA.keys()))
            del patient.marketDNA[randomKey]
        
        return patient
        
    """
    """
    Mutate individual
    """
    def Save(self, folder = ""):
        import pickle
       
        fileName = BOT_SAVE_FOLDER + self.name + '.bot'
        saveFile = open(fileName, 'wb')
        pickle.dump(self, saveFile)
        saveFile.close()

        return pickle.dump(self)

    def Mutate(self, args = {}):
        
        # iterate over dna and mutate 
        for dnaTag, dna in self.dNA.items():
            for genomeID, genome in dna.genome.items():
                for geneTag, gene in genome.genome.items():
                    gene.mutate(args)

    
"""

class myTradeBot(BotRoot):
    def __init__(self, myenv="", myspecies = "myTradeBot", myargs=None):
        super().__init__(myenv=myenv, myspecies = myspecies, myargs=myargs)

        #looking for a buyer here
        self.direction = 1

        def Mutate(self):
            print ("Mutate bot...")


"""



"""Population class for evolutionary algorithms
"""
import json, requests
import logging
from tqdm import tqdm as tq
import os, sys, random, pkgutil, math
logging.basicConfig(level=logging.CRITICAL)
#insert root dir into system path for python
#config core
# sys.path.append('../')
# from configCore import *

#import root bot
from marlin_brahma.bots.bot_root import *

# import bots
from custom_bots import *
# from custom_genes import *

import os


class Population(object):
    """
    Population is the container object for all bots/indis in the game/optimisation run.
    
    Arguments:
        object {root object} -- object class
    """

    def __init__(self, parms=None, name="vx_dm", gene_args = None, version="1_0_0"):
       
        # tb = testBot()
        """
        Create the Population class. This class holds all bots in the game.

        :param parms: optimisation parameters - from config file but passed through for now, defaults to None
        :type parms: key/value pairs, optional
        :param name: name root, defaults to "vx_dm"
        :type name: str, optional
        """

        
        
        self.name = name + str(random.randint(10,10000))
        self.parms = parms
        self.feature_version = version
        #bot names in list
        self.bots = []
        #species structure tagged by name
        self.species = {}
        #logger

        #self.logger = GetBrahmaLogger("Population")
        self.gene_args = gene_args
    def __exit__(self):
        del self.logger

    #edit march
    #kill bots not making decisions -> 0 fitness
    
    def KillDeadWood(self, tags = []):
        number = len(tags)
        splice_number = math.floor(number/2)
        for tag in tags[0:splice_number]:
            if tag in self.species:
                del (self.species[tag])
                self.bots.remove(tag)
   
    #edit march
    def Repopulate(self, species = "BotRoot", args = None):
        population_size = len(self.species)
        delta = self.parms['population_size'] - population_size
        
        for i in range(delta):
            self.CreateBot(species = species, args = args) 
            
        for i in range(2):
            self.CreateBot(species = species, args = args) 
           
        
    def Populate(self, species = "BotRoot", args = None):
        logging.debug('Building inside brahma pop ')
        if globals().get("AcousticBot") is not None:
            print ("We have bot creation dust! ")
            
       
        import pickle
        import os
        
        if args == None:
            # print ("new bots")
            if species != "tribal":
                # if DEBUG:
                #     print ("Size: " + str(self.parms['population_size']))
                # print ("Size: " + str(self.parms['population_size']))
                population_size = self.parms['population_size']
                print (self.gene_args)
                for i in tq(range(0, int(self.parms['population_size']))):
                    
                    logging.debug(f'building {i} of {population_size}')
                    self.CreateBot(species = species, args = args)   
                    logging.debug(f'built {i} of {population_size}')
                    
            else:

                self.CreateBot(species = species, args = args)

            
            # print ("populated")
            # self.logger(self.show())
            #self.Show()
            
            
            
            
            
            
        if args=="Living":
            
            BOT_SAVE_FOLDER = os.path.join(os.path.expanduser('~'), 'dev', 'app', 'tutorial-make-decisions', 'saved', '')
            # load all traders in bin folder
            listOfFiles = os.listdir(BOT_SAVE_FOLDER)
            
            for i in range(0, int(self.parms['population_size'])):
                
                #load saved bots
                filename = random.choice(listOfFiles)
                
                #pop from list
                indx = listOfFiles.index(filename)
                listOfFiles.pop(indx)
                
                try:
                    print (BOT_SAVE_FOLDER + filename)
                    pkl_binary = open(BOT_SAVE_FOLDER + filename , 'rb')
                   
                except:
                    print ("warning, bot not found")
                    exit()
                    continue
            
                try:
                    _bot = pickle.load(pkl_binary)
                    self.ResetTrader(_bot)
                    self.bots.append(_bot.name)
                    self.species[_bot.name] = _bot
                except Exception as e:
                    print (e)
                    
                
                # print (_bot.name)
                # print (_bot)
                
                
                    
    def ResetTrader(self,  trader):
        for dnaTag, dna in trader.dNA.items():
            dna.expressionTable = {}
            

    def test(self):
        pass


    # def LoadBot():
        

    def CreateBot(self, species = None, args = None):
        self.botStr = {}
        self.botStr[species] = eval(species)
        
        logging.debug(f'creating bot')
        #build the bot here - general
        
        bot_tmp = self.botStr[species](self.parms["env"], myspecies = species, myargs = args, version=self.feature_version)
        logging.debug(f'tmp bot build')
        bot_tmp.BuildBot(parms=self.parms, gene_args = self.gene_args)
        logging.debug(f'tmp bot build 2')
        #bot_tmp.printAll()
        self.bots.append(bot_tmp.name)
        self.species[bot_tmp.name] = bot_tmp
        
        '''
        print ("Building.. " + species)
        bot_tmp = trader_template(self.Parms['market'])
        bot_tmp.build_trader(parms=self.Parms)
        self.Bots.append(bot_tmp.Name)
        self.Species[bot_tmp.Name] = bot_tmp
        bot_tmp = None
        '''

    def Show(self):
        
        for key, value in self.species.items():
            # print (value)
            value.printAll()

            
    def save_bots(self):
        for bot_id, bot in self.species.items():
            BOT_SAVE_FOLDER = "bots/"
            #serialise bot to local file.
            print (f'saving ... {bot_id}')
            bot.save(save_folder=BOT_SAVE_FOLDER)
            self.recordWinningBot(bot)
            
    
    def recordWinningBot(self, bot):
        """Record optimisation winners to db

        Arguments:
            botlist {[type]} -- [description]
        """
        #logger.info('Winning bots recorded')
        botStr = bot
        botID = bot.name
        #--build data for posting
        dataSend = {}
        dataSend["action"] = "record_bot"
        dataSend["user"] = self.parms['uid']
        dataSend["botID"] = botID
        dataSend["botStructure"] = botStr.printStr()
        dataSend["market"] = self.parms['env']
        dataSend["direction"] = ""
        dataSend["optimisationID"] = ""
        dataSend["parent"] = "Eve"
        dataSend["scope"] = "global"
       
        
        dataSendJSON = json.dumps(dataSend)
      
        #--post data
        try:
            API_ENDPOINT = "https://www.vixencapital.com/api/optimisation/"
            try:
                r = requests.post(url = API_ENDPOINT, data = dataSendJSON)
                response = r.text
            except:
                print("request error")

        except:
            #logger.critical('Error recording winning bot')
            pass


        #print (response)
    


if __name__ == "__main__":
    pass
    '''
    pop = Population(parms={'PopulationSize' : '100'}, name = "demo")
    pop.populate('trader')
    '''

